@@ -1,20 +1,25 @@
 # mypy: ignore-errors
 from __future__ import annotations
 
-from typing import Union
+from typing import Union, Any, Callable
 from collections import Counter
 from contextlib import nullcontext
 
-
 from graphviz import Digraph
 from torch.nn.modules import Identity
 
 from .computation_node import NodeContainer
 from .computation_node import TensorNode, ModuleNode, FunctionNode
-from .utils import updated_dict
+from .utils import updated_dict, assert_input_type
 
 COMPUTATION_NODES = Union[TensorNode, ModuleNode, FunctionNode]
 
+node2color = {
+    TensorNode: "lightyellow",
+    ModuleNode: "darkseagreen1",
+    FunctionNode: "aliceblue",
+}
+
 # TODO: Currently, we only use directed graphviz graph since DNN are
 # graphs except for e.g. graph neural network (GNN). Experiment on GNN
 # and see if undirected graphviz graph can be used to represent GNNs
@@ -22,12 +27,6 @@ COMPUTATION_NODES = Union[TensorNode, ModuleNode, FunctionNode]
 # TODO: change api of to include also function calls, not only pytorch models
 # so, keep the api here as general as possible
 
-node2color = {
-    TensorNode: "lightyellow",
-    ModuleNode: "darkseagreen1",
-    FunctionNode: "aliceblue",
-}
-
 
 class ComputationGraph:
     '''A class to represent Computational graph and visualization of pytorch model
@@ -82,7 +81,7 @@ class ComputationGraph:
         self.roll = roll
         self.depth = depth
 
-        # specs for html table
+        # specs for html table, needed for node labels
         self.html_config = {
             'border': 0,
             'cell_border': 1,
@@ -93,7 +92,7 @@ class ComputationGraph:
         }
         self.reset_graph_history()
 
-    def reset_graph_history(self):
+    def reset_graph_history(self) -> None:
         '''Resets to id config to the setting of empty visual graph
         needed for getting reproducible/deterministic node name and
         graphviz graphs. This is especially important for output tests
@@ -102,7 +101,7 @@ class ComputationGraph:
         self.running_node_id: int = 0
         self.running_subgraph_id: int = 0
         self.id_dict: dict[str, int] = {}
-        self.added_nodes: set[int] = set()
+        self.node_set: set[int] = set()
         self.edge_list: list[tuple[COMPUTATION_NODES, COMPUTATION_NODES]] = []
 
         # module node  to capture whole graph
@@ -118,47 +117,45 @@ class ComputationGraph:
         for root_node in self.root_container:
             root_node.context = self.node_hierarchy[main_container_module]
 
-    def fill_visual_graph(self):
+    def fill_visual_graph(self) -> None:
         '''Fills the graphviz graph with desired nodes and edges.'''
 
         self.render_nodes()
         self.render_edges()
         self.resize_graph()
 
-    def render_nodes(
-        self,
-    ):
+    def render_nodes(self) -> None:
         kwargs = {
             'cur_node': self.node_hierarchy,
             'subgraph': None,
         }
         self.traverse_graph(self.collect_graph, **kwargs)
 
-    def render_edges(self):
-        '''Adds all edges in self.edge_list to
+    def render_edges(self) -> None:
+        '''Records all edges in self.edge_list to
         the graphviz graph using node ids from edge_list'''
-        counter_edge = {}
+        edge_counter: dict[tuple[int, int], int] = {}
         for tail, head in self.edge_list:
-            tail_id, head_id = self.id_dict[tail.node_id], self.id_dict[head.node_id]
-            counter_edge[(tail_id, head_id)] = (
-                counter_edge.get((tail_id, head_id), 0) + 1
-            )
-            self.add_edge(
-                tail_id, head_id, counter_edge[(tail_id, head_id)]
-            )
+            edge_id = self.id_dict[tail.node_id], self.id_dict[head.node_id]
+            edge_counter[edge_id] = edge_counter.get(edge_id, 0) + 1
+            self.add_edge(edge_id, edge_counter[edge_id])
 
     def traverse_graph(
-        self, action_fn, **kwargs
-    ):
+        self, action_fn: Callable[..., None], **kwargs: Any
+    ) -> None:
         cur_node = kwargs['cur_node']
         cur_subgraph = (
             self.visual_graph if kwargs['subgraph'] is None else kwargs['subgraph']
         )
+        assert_input_type(
+            'traverse_graph', (TensorNode, ModuleNode, FunctionNode, dict), cur_node
+        )
         if isinstance(cur_node, (TensorNode, ModuleNode, FunctionNode)):
             if cur_node.depth <= self.depth:
                 action_fn(**kwargs)
+            return
 
-        elif isinstance(cur_node, dict):
+        if isinstance(cur_node, dict):
             k, v = list(cur_node.items())[0]
             new_kwargs = updated_dict(kwargs, 'cur_node', k)
             if k.depth <= self.depth and k.depth >= 0:
@@ -187,19 +184,17 @@ class ComputationGraph:
                 for g in v:
                     new_kwargs = updated_dict(new_kwargs, 'cur_node', g)
                     self.traverse_graph(action_fn, **new_kwargs)
-        else:
-            raise ValueError('this should not be reached')
 
-    def collect_graph(
-        self, **kwargs,
-    ) -> None:
+    def collect_graph(self, **kwargs: Any) -> None:
         '''Adds edges and nodes with appropriate node name/id (so it respects
         properties e.g. if rolled recursive nodes are given the same node name
         in graphviz graph)'''
 
         cur_node = kwargs['cur_node']
         # if tensor node is traced, dont repeat collecting
-        if id(cur_node) in self.added_nodes:
+        # if node is isolated, dont record it
+        is_isolated = cur_node.is_root() and cur_node.is_leaf()
+        if id(cur_node) in self.node_set or is_isolated:
             return
 
         self.check_node(cur_node)
@@ -246,12 +241,12 @@ class ComputationGraph:
         # # output node
         # visible tensor and non-input tensor nodes
         if is_cur_visible and not cur_node.is_root():
-            assert not isinstance(tail_node, TensorNode) or tail_node.is_root()
+            assert not isinstance(tail_node, TensorNode) or tail_node.is_root(), (
+                "get_tail_node function returned inconsistent Node, please report this"
+            )
             self.edge_list.append((tail_node, cur_node))
 
-    def rollify(
-        self, cur_node: ModuleNode | FunctionNode
-    ):
+    def rollify(self, cur_node: ModuleNode | FunctionNode) -> None:
         '''Rolls computational graph by identifying recursively used
         Modules. This is done by giving the same id for nodes that are
         recursively used.
@@ -267,10 +262,14 @@ class ComputationGraph:
         output_id = get_output_id(head_node)
         cur_node.set_node_id(output_id=output_id)
 
-    def is_node_visible(self, compute_node):
+    def is_node_visible(self, compute_node: COMPUTATION_NODES) -> bool:
         '''Returns True if node should be displayed on the visual
         graph. Otherwise False'''
 
+        assert_input_type(
+            'is_node_visible', (TensorNode, ModuleNode, FunctionNode,), compute_node
+        )
+
         if compute_node.name == 'empty-pass':
             return False
 
@@ -283,7 +282,7 @@ class ComputationGraph:
             )
             return is_visible
 
-        if isinstance(compute_node, TensorNode):
+        else:
             if compute_node.main_node.depth < 0 or compute_node.is_aux:
                 return False
 
@@ -297,11 +296,7 @@ class ComputationGraph:
 
             return is_visible
 
-        raise ValueError(
-            'Only Computation Nodes are allowed for this function as an input'
-        )
-
-    def get_tail_node(self, _tensor_node: TensorNode):
+    def get_tail_node(self, _tensor_node: TensorNode) -> COMPUTATION_NODES:
 
         tensor_node = _tensor_node.main_node if _tensor_node.is_aux else _tensor_node
 
@@ -319,7 +314,7 @@ class ComputationGraph:
             if depth >= self.depth:
                 break
 
-        module_depth = depth-1
+        module_depth = depth - 1
         # if returned by container module and hide_module_functions
         if (
             isinstance(current_parent_h[depth], FunctionNode) and
@@ -344,15 +339,19 @@ class ComputationGraph:
         return tail_node
 
     def add_edge(
-        self, tail_id: int, head_id: int, edg_cnt: int
+        self, edge_ids: tuple[int, int], edg_cnt: int
     ) -> None:
 
+        tail_id, head_id = edge_ids
         label = None if edg_cnt == 1 else f' x{edg_cnt}'
         self.visual_graph.edge(f'{tail_id}', f'{head_id}', label=label)
 
     def add_node(
         self, node: COMPUTATION_NODES, subgraph: Digraph | None = None
     ) -> None:
+        '''Adds node to the graphviz with correct id, label and color
+        settings. Updates state of running_node_id if node is not
+        identified before.'''
         if node.node_id not in self.id_dict:
             self.id_dict[node.node_id] = self.running_node_id
             self.running_node_id += 1
@@ -364,9 +363,12 @@ class ComputationGraph:
         subgraph.node(
             name=f'{self.id_dict[node.node_id]}', label=label, fillcolor=node_color,
         )
-        self.added_nodes.add(id(node))
+        self.node_set.add(id(node))
 
     def get_node_label(self, node: COMPUTATION_NODES) -> str:
+        '''Returns html-like format for the label of node. This html-like
+        label is based on Graphviz API for html-like format. For setting of node label
+        it uses graph config and html_config.'''
         input_str = 'input'
         output_str = 'output'
         border = self.html_config['border']
@@ -404,7 +406,12 @@ class ComputationGraph:
                     </TABLE>>'''
         return label
 
-    def resize_graph(self, scale=1.0, size_per_element=0.3, min_size=12):
+    def resize_graph(
+        self,
+        scale: float = 1.0,
+        size_per_element: float = 0.3,
+        min_size: float = 12
+    ) -> None:
         """Resize the graph according to how much content it contains.
         Modify the graph in place. Default values are subject to change,
         so far they seem to work fine.
@@ -422,13 +429,10 @@ class ComputationGraph:
     ) -> str:
         return node2color[type(node)]
 
-    def check_node(self, node):
+    def check_node(self, node: COMPUTATION_NODES) -> None:
         assert node.node_id != 'null', f'wrong id {node} {type(node)}'
-        assert '-' not in node.node_id, 'No repetition of node recording'
-        assert not node.is_leaf() or not node.is_root(), (
-            f'isolated node! {node}'
-        )
-        assert node.depth <= self.depth, f"exceeds depth limit, {node}"
+        assert '-' not in node.node_id, 'No repetition of node recording is allowed'
+        assert node.depth <= self.depth, f"Exceeds display depth limit, {node}"
         assert (
             sum(1 for _ in node.parents) in [0, 1] or not isinstance(node, TensorNode)
         ), (
@@ -436,7 +440,7 @@ class ComputationGraph:
         )
 
 
-def compact_list_repr(x: list):
+def compact_list_repr(x: list[Any]) -> str:
     '''returns more compact representation of list with
     repeated elements. This is useful for e.g. output of transformer/rnn
     models where hidden state outputs shapes is repetation of one hidden unit
@@ -455,20 +459,16 @@ def compact_list_repr(x: list):
     return x_repr[:-2]
 
 
-def get_output_id(head_node: COMPUTATION_NODES) -> str | int:
+def get_output_id(head_node: COMPUTATION_NODES) -> str:
     ''' This returns id of output to get correct id.
     This is used to identify the recursively used modules.
     Identification relation is as follows:
         ModuleNodes => by id of nn.Module object
-        Parameterless Modules => by id Node object
+        Parameterless ModulesNodes => by id of nn.Module object
         FunctionNodes => by id of Node object
     '''
     if isinstance(head_node, ModuleNode):
-        if head_node.is_activation:
-            # TODO: try also compute_unit_id
-            output_id = head_node.compute_unit_id
-        else:
-            output_id = head_node.compute_unit_id
+        output_id = str(head_node.compute_unit_id)
     else:
         output_id = head_node.node_id
 
